%  slirputils.sl:	SLIRP utility routines {{{
%
%  This file is part of SLIRP, the (Sl)ang (I)nte(r)face (P)ackage.
%
%  Copyright (c) 2003-2004 Massachusetts Institute of Technology
%  Copyright (C) 2002 Michael S. Noble <mnoble@space.mit.edu>
%
%  This software was partially developed by the MIT Center for Space
%  Research under contract SV1-61010 from the Smithsonian Institution.
%  
%  Permission to use, copy, modify, distribute, and sell this software
%  and its documentation for any purpose is hereby granted without fee,
%  provided that the above copyright notice appear in all copies and
%  that both that copyright notice and this permission notice appear in
%  the supporting documentation, and that the name of the Massachusetts
%  Institute of Technology not be used in advertising or publicity
%  pertaining to distribution of the software without specific, written
%  prior permission.  The Massachusetts Institute of Technology makes
%  no representations about the suitability of this software for any
%  purpose.  It is provided "as is" without express or implied warranty.
%  
%  THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY DISCLAIMS ALL WARRANTIES
%  WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
%  MERCHANTABILITY AND FITNESS.  IN NO EVENT SHALL THE MASSACHUSETTS
%  INSTITUTE OF TECHNOLOGY BE LIABLE FOR ANY SPECIAL, INDIRECT OR
%  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
%  OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
%  NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
%  WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. }}}

% Front Matter: type, variable and forward declarations {{{ 

typedef struct {
   name,
   int_consts,
   double_consts,
   string_consts,
   intrin_vars,
   intrin_arrays,
   inlines,		% general inlined code (for body of translation unit)
   inlines_init,	% inlined code for module initialization fragment
   functions,
   prototypes,
   unresolved_macros
} Interface;

typedef struct {
   name,
   ancestors,
   constructors,
   abstract,
   nesting,
   interface,		% class-scoped public interface
} Class;

variable SlirpController = struct {

  gendir,		% directory into which code will be generated
  genstem,		% path-qualified stem of filenames for generated code

  filename,		% name of input file currently being parsed
  srcfp,		% FILE pointer to same
  gluefp,		% target for emission of generated wrapper code 

  debug,
  autotype,
  cfront,		% indicates whether pure C wrappers should be generated
  emit_usage,
  usage_strings,
  assume_const_strings,
  always_pop_args,

  argc,
  tablevel,		% pretty printing mechanism
  status,		% status returned to O/S upon termination
  decl,			% current declaration being constructed/examined/parsed
  inlined,

  funcprefix,		% generate code for all funcs which match this regexp
  renames,		% maps C function names to different S-Lang names

  interface,		% currently active interface (global or class namespace)
  classes,
  cplusplus,		% indicates whether current file is known to be C++
  dispatch_table,
  in_public_interface,	% indicated when parsing global or public class content
  types,
  opaque_types,
  num_reserved_types,

  modname,		% name of module for which code is being generated
  include_paths,	% -I options specified at invocation
  lib_paths,		% -L options specified at invocation
  lib_names,		% -l options specified at invocation
  ldflags,		% -ldflags options specified at invocation
  header_names,		% name(s) of header file(s), 2b #include-d w/in module
  header_paths,		% path(s) to them (supplemented by -I include_paths)

  map_scalars_to_refs,	% indicates whether generated code will implicitly
     			% map scalars to references (e.g. to satisfy FORTRAN
			% pass-by-reference semantics, w/out requiring the 
			% S-Lang user to explicitly use & at call time)

  fortran,		% indicates that current input file is fortran; for
			% example, removal of trailing underscore from the
			% function name (at most only two will be ellided)
			% also implies map_scalars_to_refs = 1

  have_f2c,		% indicates that f2c was detected on build system
  have_fortran,		% indicates at least one input file was fortran
  have_refs,
  have_null_term_str_arrays,

  print_interface,
  genstubs,
  geninit,		% indicates whether module init fragment s/b emitted
  listopaques,		% indicates whether list of opaque types s/b emitted
  strippingComments,
  genmakef,		% indicates whether sample Makefile will be generated

  root,			% root of tree from which SLIRP is running
  version,		% current SLIRP version
  rcfile,		% name of SLIRP resource file in use (if any)


  tmap_infile,		% name of file to which typemap will be dumped
  tmap_outfile,		% name of file form which typemap will be read 
  wrapper_prefix	% initial part of generated function name
};

public variable SC = @SlirpController,	% Slirp controller instance
		EMPTY = "",
		VOID = "void",
		DEFS_COMMENT_CHAR =';';

static variable ch = "",		% last character read from file
		line = "";		% last line read from file

define finalize_ignore_list();
define slirp_include_rc();
define list_opaque_types();
define generate_prototype();
define get_token();
define get_line();
define swallow_rest_of_C_block();
define swallow_struct_or_union();
variable slirp_init_cpp_callback = NULL;
variable slirp_init_cfront_callback;
variable slirp_init_genstubs_callback;

define dprintf()
{
   variable args = __pop_args(_NARGS);
   () = fprintf(stderr,__push_args(args));
   () = fprintf(stderr,"\n");
   () = fflush(stderr);
}

define tprintf()
{ 
   variable args = __pop_args(_NARGS);
   loop(SC.tablevel)
      () = printf("\t");
   () = printf(__push_args(args));
   () = printf("\n");
   () = fflush(stdout);
}

public define Empty_Assoc_Array()	{ return  Assoc_Type[Any_Type,NULL]; }

SC.debug = 0;
SC.autotype = 1;
SC.opaque_types = Struct_Type[0];
SC.types = Empty_Assoc_Array;
SC.funcprefix = "^[a-zA-Z]+";
SC.wrapper_prefix = "sl_";
SC.cfront = 0;
SC.have_f2c = 0;
variable accepts_null_args = Empty_Assoc_Array;

% }}}

% Linked lists {{{

variable List = struct { head, tail }; 
variable List_Elem = struct { name, value, next, prev };

define _list_append(list,node)	% O(1) insertion
{
   if (list.head == NULL) {
	list.head = node;
	list.tail = node;
   }
   else {
	node.prev = list.tail;
	list.tail.next = node;
	list.tail = node;
   }
}

define _list_delete(list,node)	% O(1) removal
{
   if (node.prev != NULL)
	node.prev.next = node.next;
   else
	list.head = node.next;

   if (node.next == NULL) list.tail = node.prev;
}

define list_node_new(name, value)
{
   variable node = @List_Elem;
   node.name = name;
   node.value = value;
   return node;
}

define add_const(hash, name, value)
{
   hash[SC.interface.name + name] = value;
}
% }}}

% File I/O and tokenization {{{

define abort()
{
   variable args = __pop_args(_NARGS);
   dprintf(__push_args(args));
   exit(1);
}

define emit()
{
   variable args = __pop_args(_NARGS);
   () = fprintf(SC.gluefp, __push_args(args));
}


define emit_inlines(inlines)
{
   foreach(inlines) { variable il = (); emit("%s",il); }
}

define eof_error(f)
{
   variable msg = sprintf("unexpected EOF in %S, near line <%S> of file <%S>",
	 		f, line, SC.filename);
   if (SC.debug)
	verror(msg);
   else 
	abort(msg);
}

public define get_rest_of_line(fp)
{
   variable l = get_line(fp,0);
   while (l[-1] == '\\')		% support C-style line continuation
	l = l + get_line(fp,0);
   return l;
}

define swallow_C_style_comment(fp)
{
    variable current_line = get_rest_of_line(fp);

    while(current_line[[-2:-1]] != "*/") {
	current_line = get_line(fp,1);
	if (current_line == NULL)
	    break;
    }
}

define backup(fp, n)
{
   if (fseek(fp,-n,SEEK_CUR) == -1)
	verror("fseek error, near line: %S",line);
}

define swallow_to(fp,delim)
{
   forever { 
	line = get_rest_of_line(fp);
	variable pos = is_substr(line,delim);
	if (pos) {
	   variable l = strlen(line);
	   if (pos < l)
		backup(fp, strlen(line) - pos - strlen(delim) + 2);
	   break;
	}
   }
}

define get_char(fp)
{
   if ( fread (&ch, Char_Type, 1, fp) == -1)
	eof_error(_function_name);
   return ch;
}

define next_char(fp)  {variable ch = get_char(fp); backup(fp,1); return ch;}

static define strip_comments(s, fp, is_token)
{
   if (SC.strippingComments)
	return s;

   SC.strippingComments = 1;

   forever {
	variable bc = is_substr(s,"/*");
	if (bc) {
	   variable ec = string_match(s,"*\\/",bc);
	   if (ec) {
#ifeval _slang_version < 20000
		ec = ec + bc - 1;		% work around slang1 bug
#endif
		s = strtrim(strcat(substr(s,1,bc-1),substr(s,ec+2,-1)));
	   }
	   else {
		s = strtrim(substr(s,1,bc-1));
		swallow_to(fp,"*/");
	   }
	}
	else {
	   bc = is_substr(s,"//"); 	% Handle C++ comments, too, even tho
	   if (bc) {			% they don't belong in C code proper
		if (andelse {is_token} {next_char(fp) != '\n'})
		() = get_rest_of_line(fp);
		s = strtrim(substr(s,1,bc-1));
	   }
	   break;
	}
   }

   SC.strippingComments = 0;
   return s;
}

public define get_line(fp, okNULL)
{
   variable l;

   do {
	if (fgets(&l,fp) == -1) {
	   if (okNULL)
		return NULL;
	   else
		eof_error(_function_name);
	}

	% Compress whitespace (only kind of whitespace remaining is a space)
	l = strcompress(l," \t\r\n\f");

	% Ignore empty lines and Scheme comments (which s/b only w/in .defs)
	if (andelse {l != EMPTY} {l[0] != DEFS_COMMENT_CHAR})
	   l = strip_comments(l, fp, 0);

   } while (l == EMPTY);		% return only non-blank lines

   return l;
}

define malformed_error(thing)
{
   verror("malformed %s, near line: <%S>\n"+
	  "within file <%S>\n"+
	  "last character read: <%c>", thing, line, SC.filename, ch);
}

define is_white(c) {
   return (orelse {c == ' '} {c == '\t'} {c == '\n'} {c == ''} );
}

define eat_white(fp)
{
   do {
   	variable ch = get_char(fp);

	% Treat comments as whitespace (we support both C and C++ style)
	if (ch == '/') {

	   switch (get_char(fp))
		{ case '*' :	swallow_C_style_comment(fp);
		   		ch = get_char(fp);
		}
		{ case '/' :	() = get_rest_of_line(fp);
		   		ch = get_char(fp);
		}
		{ backup(fp, 1); }
	}

   } while (is_white(ch));

   backup(fp, 1);
}

public variable Single_Char_Tokens = Assoc_Type[Char_Type, 0];
public define add_single_char_tokens(space_delimited_string_of_chars)
{
   foreach (strtok(space_delimited_string_of_chars, " \t")) {
	variable s = ();
	Single_Char_Tokens[s] = int(s);
   }
}
add_single_char_tokens("{ } :");


public define get_token(fp)
{
   % Simple tokenizer, looking mainly for whitespace-delimited
   % sequences.  Comments are automatically stripped out.

   variable ch, token = EMPTY;

   do {
	if ( fread (&ch, Char_Type, 1, fp) == -1)	% eat leading whitesp
	   return token;
   } while (is_white(ch));

   backup(fp, 1);

   while (1) {

 	if ( fread (&ch, Char_Type, 1, fp) == -1)
	   break;

	if (is_white(ch))
	   break;

	if (Single_Char_Tokens[char(ch)]) {
	   if (token != EMPTY) { backup(fp, 1); break; }
	   return char(ch);
	}
	else if (ch == '#')		% skip space after #, so as to combine
	   eat_white(fp);		% # and next literal into single token

	token = strcat(token,char(ch));
   }

   token = strip_comments(token, fp, 1);
   if (token == EMPTY)
	return get_token(fp);

   return token;
}

define unget_token(fp, tok)
{
   variable l = strlen(tok);
   if (l > 1) l++;
   backup(fp, l);
}

define next_token(fp) {
   variable prev = ftell(fp);
   variable t = get_token(fp);
   () = fseek(fp, prev - ftell(fp), SEEK_CUR);
   return t;
}

define open_block(fp,block_open_delimiter)
{
   eat_white(fp);
   if (get_char(fp) != block_open_delimiter) {
      backup(fp, 1);
      return 0;
   }
   return 1;
}

define open_C_block(fp)	   { return open_block(fp,'{'); }

define close_block(fp,block_close_delimiter)
{
   eat_white(fp);
   if (get_char(fp) != block_close_delimiter)
	malformed_error("block close");
}

define close_C_block(fp) { close_block(fp,'}'); }

define swallow_rest_of_C_block(fp)
{
   do {
	variable token = get_token(fp);
	if (token == "{")
	   swallow_rest_of_C_block(fp);
   } while (token != "}");
}

define swallow_C_block(fp)
{
   swallow_to(fp, "{");
   swallow_rest_of_C_block(fp);
}

define swallow_struct_or_union(fp)
{
   !if (open_C_block(fp)) {
	() = get_token(fp);			% ignore any identifier
	!if (open_C_block(fp)) {
	   swallow_to(fp,";");			% simple typedef of other struct
	   return;
	}
   }

   while (1) {
	variable token = get_token(fp);
	switch(token)
	{ case "struct" or case "union" : swallow_struct_or_union(fp); }
	{ case "}": break; }
   }
   swallow_to(fp,";");
}

define extract_enumerator(fp)
{
   eat_white(fp);
   variable name = "", need_value = 0, need_name = 1;

   while (1) {

	() = get_char(fp);
	switch(ch)
	   { case ',' or case '}': backup(fp, 1); return name;	}
	   { case '=': need_name = 0; need_value = 1;		}
	   { case '\n': !if (need_value) return name;		}
           { case ' ' or case '\t' : continue;			}
           { case '#' : get_rest_of_line(fp);			}
	   { case '/': 
	      	if (get_char(fp) == '*') {
		   swallow_C_style_comment(fp);
		   return name;
		}
		backup(fp, 1);	% else clause
		continue;
	   }
	   {
	   	if (need_name)
		   name = strcat(name,char(ch));
		else
		   need_value = 0;
	   }
   }
}

define more_enum_values(fp) {

   eat_white(fp);
   if (get_char(fp) == ',')
	return 1;

   backup(fp, 1);
   return 0;
}

define get_match(s, which)
{ 
   variable pos, len;
   (pos, len) = string_match_nth(which);
   if (len == 0) return EMPTY;
   return s[[pos:pos+len-1]];
}

define get_matchc(s, which) { return strcompress(get_match(s,which)," \t"); }

define group_pointer_qualifiers_with_type(decl)
{
   % Assist parsing of identifier declarations, by ensuring that pointer
   % or reference qualifiers are grouped with the identifier type (not name)

   % First convert array specifiers of indeterminate size to generic pointers
   forever  {
	variable pos, len;
	pos = string_match(decl,"\\([_A-Za-z]+\\)[ \t]*\\[[ \t]*\\]\\(.*\\)",1);
	!if (pos) break;
	decl = sprintf("%s* %s%s", decl[[:pos-2]], get_match(decl, 1),
						  get_match(decl, 2));
   }

   (decl,) = strreplace(decl, " &", "&", 99);
   (decl,) = strreplace(decl, "\t&", "&", 99);
   (decl,) = strreplace(decl, " *", "*", 99);
   (decl,) = strreplace(decl, "\t*", "*", 99);
   forever  {
	pos = string_match(decl, "[*&][_A-Za-z]+", 1);
	!if (pos) break;
	decl = sprintf("%s %s", decl[[:pos-1]], decl[[pos:]]);
   }
   return decl;
}

define trim_around(string, pattern)
{
   variable patterns = [ sprintf("\\([ \t]+\\)\\(\\%s\\)",pattern), 
   			 sprintf("\\(\\%s\\)\\([ \t]+\\)",pattern) ];

   foreach (patterns) {
	variable pat = (), substr;
	while (string_match(string, pat, 1)) {
		substr = get_match(string, 0);
		(string, ) = strreplace(string, substr, pattern, 1);
	}
   }
   return string;
}
% }}}

% Interface manipulators {{{
define save_interface(interface, initialize)
{
   variable saved_interface = @interface;

   if (initialize) {
	interface.name = EMPTY;
	interface.int_consts = Empty_Assoc_Array;
	interface.double_consts = Empty_Assoc_Array;
	interface.string_consts = Empty_Assoc_Array;
	interface.intrin_vars = Empty_Assoc_Array;
	interface.inlines = String_Type[0];
	interface.inlines_init = String_Type[0];
	% intrin_arrays intentionally kept NULL
	interface.functions  = Empty_Assoc_Array;
	interface.prototypes = Empty_Assoc_Array;
	interface.unresolved_macros = Empty_Assoc_Array;
   }

   return saved_interface;
}

define restore_interface(from, to)
{
   to.name = from.name;
   to.int_consts = from.int_consts;
   to.double_consts = from.double_consts;
   to.string_consts = from.string_consts;
   to.intrin_vars = from.intrin_vars;
   to.functions  = from.functions;
   to.inlines = from.inlines;
   to.inlines_init = from.inlines_init;
   to.prototypes = from.prototypes;
   to.unresolved_macros = from.unresolved_macros;
}

static define print_consts(kind, consts)
{
   variable n = 0;
   if (length(consts)) {
	tprintf("Public %s:", kind);

	foreach(consts) using ("keys") {
	   variable name = ();
	   tprintf("   %S",name);
	   n++;
	}
   }
   return n;
}

define print_interface(interface, description) % {{{
{
   tprintf(""); tprintf(description); SC.tablevel++;

   SC.cfront = 1;				% reflect C types in arglists
   variable thing, numfuncs = 0, overload;
   tprintf("Public Wrapper Functions:");
   foreach (interface.functions) {
	thing = ();
	thing.result.lname = EMPTY;
	thing.pop_args = 0;			% ensure arglists are non-void
	!if (length(thing.overloads)) {
	   tprintf("   %s;", generate_prototype(thing));
	   numfuncs++;
	}
	else {
	   variable overloads = 0;
	   foreach(thing.overloads) {
		variable ol = ();
		ol.pop_args = 0;		% ensure arglists are non-void
		ol.wname = ol.slname + string(overloads); 
		ol.result.lname = EMPTY;
		tprintf("   %s;", generate_prototype(ol));
		numfuncs++;
		overloads++;
	   }
	}
   }

   variable numvars = print_consts("/ Intrinsic Variables",
	 						interface.intrin_vars);
   variable numconsts = print_consts("String Consts", interface.string_consts);
   numconsts += print_consts("Integer Consts", interface.int_consts);
   numconsts += print_consts("Double Consts", interface.double_consts);

   SC.tablevel--; tprintf("");

   return (numfuncs, numvars, numconsts);
} % }}}

define print_public_interface() % {{{
{
   variable nfuncs, nvars, nconsts;
   (nfuncs, nvars, nconsts) = print_interface(SC.interface,"Global Namespace");
   foreach(SC.classes) using ("values") {
	variable class = ();
	print_interface(class.interface, "Class: " + class.name);
	nconsts += ();
	nvars   += ();
	nfuncs  += ();
   }
   tprintf("Code Generation Summary:\n"+
	"\tFunctions (includes overloads, excludes destructors): %d\n"+
	"\tVariables : %d\n"+
	"\tConstants : %d\n\n",nfuncs, nvars, nconsts);

} % }}}

private define generate_stub(funcmap, namespace) % {{{
{
   if (funcmap.inlined) return;

   variable result = funcmap.result, body;
   if (result.ltype != VOID)
	body = sprintf("return (%s%s) 0;", result.const, result.ltype);
   else
	body = EMPTY;
   result.lname = EMPTY;
   funcmap.wname = namespace + funcmap.name;
   emit("%s { %s }\n", generate_prototype(funcmap), body);
} % }}}

define generate_stubs(interface, namespace) % {{{
{
   variable thing, overload;
   foreach (interface.functions) {
	thing = ();
	thing.pop_args = 0;			% ensure arglists are non-void
	!if (length(thing.overloads)) 
	   generate_stub(thing, namespace);
	else {
	   foreach(thing.overloads) {
		variable ol = ();
		ol.pop_args = 0;		% ensure arglists are non-void
		generate_stub(ol, namespace);
	   }
	}
   }
} % }}}

define nullify_struct(s) % {{{
{
  array_map(Void_Type, &set_struct_field, s, get_struct_field_names(s), [NULL]);
} % }}}

define struct_map(type, struct_arr, field) % {{{
{
   !if (length(struct_arr)) return (@Array_Type)(type, 0);
   return array_map(type, &get_struct_field, struct_arr, field);
} % }}}

SC.interface = @Interface;
() = save_interface(SC.interface, 1);

static variable Wrapper_Semantics_Fields = ["map_scalars_to_refs", "fortran"];
define get_wrapper_semantics()
{
   variable s = (@Struct_Type) (Wrapper_Semantics_Fields);
   foreach(Wrapper_Semantics_Fields) {
	variable field = ();
	set_struct_field(s, field, get_struct_field(SC, field));
   }
   return s;
}

define set_wrapper_semantics()
{
   variable n, sem = (@Struct_Type) (Wrapper_Semantics_Fields);
   switch(_NARGS)
   { case 1: n = (); n = _push_struct_field_values(n); _stk_reverse(n); }
   { n = _NARGS; }
   variable values = __pop_args(n);
   set_struct_fields(sem, __push_args(values));
   foreach(Wrapper_Semantics_Fields) {
	variable field = ();
	variable value = get_struct_field(sem, field);
	if (value == NULL) break;
	set_struct_field(SC, field, value);
   }
}

% }}}

% Command line usage help {{{

define emit_version() {
   variable version = SC.version, revision = "0";
   if (strcmp(revision,"0"))
	version = sprintf("%S-%S", version, revision);
   printf("SLIRP Version: "+version+"\n");
   printf("SLang Library Version: "+_slang_version_string+"\n");
}

define instruct(exitvalue)
{
   () = fprintf (stderr,
   "SLIRP: The (SL)ang (I)nte(R)face (P)ackage\nA S-Lang module generator "+
   "for C, C++, and FORTRAN.\nVersion: " + SC.version + "\n\n"+
   "Usage: slirp [OPTIONS]  headerfile [ headerfile ...]\n"+
   "\n-cfront             generate standalone C wrappers (.cc and .h files),"+
   "\n                    instead of S-Lang bindings, for C++ interface"+
   "\n-const_strings      treat 'char*' return values as 'const char*'"+
   "\n-g                  emit debugging output"+
   "\n-fprefix <pref>     emit code only for functions which begin with an"+
   "\n                    exact match to <pref>, instead of default regexp"+
   "\n-h -help --help     this message"+
   "\n-I  <dir>           add <dir> to search path for headers during code"+
   "\n                    generation; will also be reflected in -make output"+
   "\n-L  <dir>           add <dir> to search path at link time; for -make"+
   "\n-l  <lib>           add <lib> to library list at link time; for -make"+
   "\n-ldflags <flags>    use when -L or -l are inappropriate; for -make"+
   "\n-m  <name>          override default module name with <name>"+
   "\n-make               generate a best-effort Makefile for compiling module"+
   "\n                    (-make is also implied by -L, -l, and -ldflags)"+
   "\n-mapnames <R> <S>   (deprecated) synonym for -rename"+
   "\n-noinit             do not generate an initialization fragment"+
   "\n-nopop              avoid explicit pop of func arguments, if possible"+
   "\n                    (also turns off generation of Usage: statements)"+
   "\n-noautotype         disable mapping of unknown types to opaque ptr types"+
   "\n-otree              print hierarchical listing of all opaque types"+
   "\n                    defined by the current invocation, then exit"+
   "\n-print              print interface for code which would be generated,"+
   "\n                    but don't actually generate any code"+
   "\n-rc  <file>         load customizations from <file>, rather than from"+
   "\n                    first of ./slirprc or $SLIRPRC"+
   "\n-refscalars         support passing scalars as array/ref arguments"+
   "\n                    (default: on for FORTRAN bindings, off for C/C++)"+
   "\n-rename   <R> <S>   map C functions with prefixes matching the S-Lang"+
   "\n                    regexp R to S-Lang functions beginning with string S"+
   "\n                    (set S=NULL to map regexp prefix to empty string)"+
   "\n-stdout             emit code to stdout, instead of files"+
   "\n-stubs              generate stub (empty) implementations for input"+
   "\n                    header files, allowing the module interface to be"+
   "\n                    exercised without linking in the underlying library"+
   "\n                    or any of its dependencies"+
   "\n-tmapout <file>     save a copy of the active typemap table to <file>"+
   "\n-tmapin  <file>     evalfile() additional typemaps from <file>"+
   "\n--version           output version information"+
   "\n-v                  show verbose loading messages\n\n"+
   "Default search path: %s\n", get_slang_load_path());

   exit(exitvalue);
}				% }}}

% Initialization / Finalization {{{
define load_user_rc()
{
   %  This routine loads module-specific extensions, such as additional
   %  typemaps, macro customizations, ignored functions, etcetera.  The
   %  content will be sought in
   %
   %	SC.rcfile 		name of SLIRP resource file (a slang script)
   % or
   %	$SLIRPRC 		environment variable naming the resource file
   %
   %  If the -rc switch is not specified then SC.rcfile defaults to
   %  "./slirprc" (file in local directory).  If SC.rcfile does not
   %  exist then $SLIRPRC is checked.

   if (SC.tmap_infile != NULL)			% support loading arbitrary
	slirp_include_rc(SC.tmap_infile);	% typemaps via cmdline switch

   if (SC.rcfile == NULL) SC.rcfile = "./slirprc";
   if (stat_file (SC.rcfile) == NULL)
	SC.rcfile = getenv("SLIRPRC");

   if (andelse {SC.rcfile != NULL} {stat_file (SC.rcfile) != NULL})
	() = evalfile(SC.rcfile);

   if (SC.listopaques) {
	list_opaque_types();
	exit(0);
   }
}

define optarg()
{
   variable default = NULL;
   if (_NARGS == 1) default = ();
   if (__argc > SC.argc + 2) {
	SC.argc++;
	return __argv[SC.argc];
   }
   return default;
}

static define is_build_arg(arg, option_name, array, test)
{
   % Check compile/link flags (used mainly w/in optional emitted Makefile)
   variable option_name_len = strlen(option_name);
   if (strncmp(arg, option_name, option_name_len)) return 0;

   variable value = arg[ [option_name_len : ] ];
   if (value == EMPTY) {
	value = optarg();
	if (value == NULL) value = EMPTY;
	% Attempt to treat quoted values sensibly
	if (orelse {value[0] == '"'} {value[0] == '\''}) {
	   variable terminator = value[0];
	   while (value[-1] != terminator) 
		value = strcat(value, optarg());
	}
	if (andelse {test} {stat_file(value) == NULL})
	   value = EMPTY;
   }

   if (value != EMPTY) array[value] = EMPTY;
   return 1;
}

define process_args()
{
   SC.argc = 1;
   while (SC.argc < __argc) {

     variable arg = __argv[SC.argc];
     switch(arg)

      { is_build_arg(arg, "-I", SC.include_paths, 1) : }
      { is_build_arg(arg, "-L", SC.lib_paths, 0) : SC.genmakef = 1; }
      { is_build_arg(arg, "-ldflags", SC.ldflags, 0) : SC.genmakef = 1; }
      { is_build_arg(arg, "-l", SC.lib_names, 0) : SC.genmakef = 1; }

      { case "-cfront":

	   SC.cfront = 1;
	   SC.wrapper_prefix = "c__";
	   @slirp_init_cfront_callback;
      }
      { case "-const_strings":	SC.assume_const_strings = 1; }

      % Use an explicit string to choose which functions to
      % wrap instead of the regular expression given above
      { case "-fprefix" :

	   SC.funcprefix = "^" + optarg("");
	   if (SC.funcprefix == "^")
		() = fprintf (stderr,"-fprefix ignored, <func_prefix> "+
			      			"argument not supplied\n");
      }

      { case "-rename" or case "-mapnames" :

	   if (__argc > SC.argc+3) {
		variable rename = String_Type[2];
		rename[0] = "^" + __argv[SC.argc+1] + "\\(.*\\)";
		rename[1] = __argv[SC.argc+2];
		if (rename[1] == "NULL")
		   rename[1] = EMPTY;
		SC.argc += 2;
		SC.renames[rename[0]] = rename[1];
	   }
	   else {
		if (__argc > SC.argc+2) SC.argc++;  % skip orphaned argument
		() = fprintf (stderr,"-mapnames ignored: "+
			      	"Usage: -mapnames <regexp> <replacement>\n");
	   }
      }

      { case "-g" : SC.debug = 1; _traceback = 1; }
      { case "-h" or case "-help" or case "--help" : instruct(0); }
      { case "-otree" : SC.listopaques = 1; }
      { case "-m" : SC.modname = optarg(); }
      { case "-make" : SC.genmakef = 1; }

      % Use S-Lang arg transfer mechanism, instead of explicit arg popping
      { case "-nopop" : SC.always_pop_args = 0; SC.emit_usage = 0; }

      { case "-noinit" : SC.geninit= 0; }
      { case "-noautotype" : SC.autotype = 0; }
      { case "-print" : SC.print_interface = 1; }
      { case "-rc" : SC.rcfile = optarg(); }
      { case "-refscalars" : SC.map_scalars_to_refs = 1; }
      { case "-stdout"  : SC.genstem  = "stdout"; }
      { case "-stubs"   :
	 	SC.genstubs = 1;
		SC.wrapper_prefix = EMPTY;
		@slirp_init_genstubs_callback;
      }
      { case "-tmapin"  : SC.tmap_infile  = optarg(); }
      { case "-tmapout" : SC.tmap_outfile =  optarg(); }
      { case "--version": emit_version(); exit(0); }
      % Note that -v option is intercepted within slshext.c
      { break; }		% Anything is considered header file input

      SC.argc++;
   }
}

public define slirp_initialize()
{
   SC.version		= "1.7.6";
   SC.root = path_dirname(__FILE__);
   if (SC.root == NULL)
	SC.root = "/build/buildd/slang-slirp-1.7.6.r3";

   if (__argc < 2)
      instruct(1);

   SC.gendir		= ".";
   SC.status		= 0;
   SC.tablevel		= 0;
   SC.emit_usage	= 1;
   SC.usage_strings	= String_Type[0];
   SC.always_pop_args   = 1;
   SC.listopaques	= 0;
   SC.dispatch_table	= String_Type[0];
   SC.geninit		= 1;
   SC.genstubs		= 0;
   SC.num_reserved_types= 0;
   SC.header_names	= String_Type[0];
   SC.include_paths	= Empty_Assoc_Array;
   SC.lib_paths		= Empty_Assoc_Array;
   SC.lib_names		= Empty_Assoc_Array;
   SC.ldflags		= Empty_Assoc_Array;
   SC.renames		= Empty_Assoc_Array;
   SC.genmakef		= 0;

   SC.classes		= Empty_Assoc_Array;
   SC.in_public_interface = 1;

   SC.assume_const_strings = 0;
   SC.cplusplus = 0;
   SC.have_refs = 0;
   SC.have_fortran = 0;
   SC.have_null_term_str_arrays = 0;

   set_wrapper_semantics(0, 0);

   _traceback = -1;
   _slangtrace = 1;

   process_args();

   % Save include paths, then morph it into a searchable path list
   SC.header_paths = SC.include_paths;
   SC.include_paths = strjoin(assoc_get_keys(SC.include_paths),
						char(path_get_delimiter()));
   preproc_activate();
   require("builtins");

   if (slirp_init_cpp_callback != NULL)
	@slirp_init_cpp_callback;

   SC.have_refs = 0;		% Reset, since builtin annotations define refs
   load_user_rc();
   finalize_ignore_list;
}

public define slirp_get_version()	{ return SC.version; }
public define slirp_check_version(version)
{
   if (SC.version >= version) return;
   abort("SLIRP version required is %S, but you have %S",version,SC.version);
}

public define close_file(fp)
{
   if (fp != NULL) {
	() = fflush(fp);
	() = fclose(fp);
   }
}

public define create_output_file()
{
   variable fname, default_extension, extension, thisfp = stdout;

   if (SC.cplusplus)
      default_extension = ".cc";
   else
      default_extension = ".c";

   switch(_NARGS)
	{ case 1: fname = (); extension = default_extension; }
	{ case 2: (fname , extension) = (); }
	{ error("incorrect number/type of arguments"); }

   if (extension == NULL) extension = EMPTY;

   if (SC.genstem != "stdout") {

	fname = sprintf("%s%s%s", SC.genstem, fname, extension);
	thisfp = fopen(fname,"w+");

	if (thisfp == NULL) {
	   () = fprintf(stderr,"Warning: could not create file: "+fname+
			 "\n\t\tI will try to emit to stdout instead");
	   thisfp = stdout;
	}
   }
   return thisfp;
}

static variable _handlers = Empty_Assoc_Array;
public define init_file_processor(originalf)
{
   variable translatedf, fullpath, fp = NULL;
   SC.strippingComments = 0;

   variable absolute = path_is_absolute (originalf);
   if (absolute)
	fullpath = originalf;
   else {
	fullpath = __find_file_in_path(SC.include_paths, originalf);
	if (fullpath == NULL)
	   fullpath = originalf;		% let it fail
	else
	   absolute = path_is_absolute (fullpath);
   }

   if (stat_file (fullpath) != NULL) {

	variable file_extension = strlow(path_extname(originalf));
	variable translator = _handlers[file_extension];
	if (translator == NULL)
	   dprintf("No language handler installed for : %s\n",originalf);
	else {
	   translatedf = @translator(fullpath);
	   if (translatedf == NULL)
		dprintf("Could not handle/translate: %s\n", originalf);
	   else
		fp = fopen(translatedf,"r");
	}
   }

   if (fp != NULL) {

	SC.filename = originalf;

	% Default modulename to stem of first file encountered
	if (SC.modname == NULL)
	   SC.modname = path_sans_extname(path_basename(originalf));

	% If this file specifies a new search path, remember it
	variable header, path = path_dirname(fullpath);
	SC.header_paths[path] = EMPTY;		% assoc array culls duplicates

	% Use a plain, instead of associative, array for header names, to
	% preserve the order in which they were specified at invocation

	if (absolute)
	   header = path_basename(translatedf);
	else
	   header = translatedf;

	!if (length( where (SC.header_names == header)))
	   SC.header_names = [ SC.header_names, header ];

	if (SC.genstem == NULL)
	   SC.genstem = SC.gendir + "/" + SC.modname;
   }
   else 
	dprintf("Skipped (unable to open or parse): %S", originalf);

   return fp;
}
% }}}

% Opaque handling code {{{
define inject_file(fname)
{
   variable buf, fp = fopen( path_concat(SC.root, fname), "r");

   while (-1 != fgets(&buf, fp))
	if (-1 == fputs(buf, SC.gluefp))
	   verror("I/O Error injecting line: <%s>\n"+
		  "from file: %s into generated code", buf, fname);

   () = fclose(fp);
}

static define friendly_remove(file)
{
   if (andelse {remove(file) != 0} {errno != ENOENT})
	message("Warning: unable to remove() file: %s",file);
}

define generate_module_init(module_defines_new_opaque_types)
{
   variable mname = SC.modname, iface = SC.interface;

   emit("BEGIN_DECLS\n");

   variable vers = strtok(SC.version, ".");
   if (length(vers) != 3)
      abort("Version number should have exactly 3 components: <%S>",SC.version);

   emit("#define   SLIRP_VERSION	%d\n", 10000*integer(vers[0]) +
	 			100*integer(vers[1]) + integer(vers[2]));
   emit("SLANG_MODULE(%s);\n", mname);
   emit("int init_%s_module_ns(char *ns_name)\n{\n", mname);
   emit("   SLang_NameSpace_Type *ns = NULL;\n\n"+
	"   if (slang_abi_mismatch()) return -1;\n"+
	"   if (ns_name != NULL) {\n"+
	"	ns = SLns_create_namespace (ns_name);\n"+
	"	if (ns == NULL) return -1;\n   }\n\n");
   emit("   slirp_debug_pause((char*)\"%s\");\n\n", mname);

   emit_inlines(SC.interface.inlines_init);
   emit("\n");

   if (SC.num_reserved_types)
	emit("   if (allocate_reserved_opaque_types() == -1) return -1;\n",
	      								mname);

   if (module_defines_new_opaque_types)
	emit("   if (allocate_%s_opaque_types() == -1) return -1;\n\n",mname);

   emit("   if (");

   if (length(iface.int_consts))
	emit(" -1 == SLns_add_iconstant_table(ns,%s_IConsts,NULL) ||\n",mname);

   if (length(iface.double_consts))
	emit("\t-1 == SLns_add_dconstant_table(ns,%s_DConsts,NULL) ||\n",mname);

   if (orelse	{length(iface.string_consts)} {length(iface.intrin_vars)} ) {
	emit("\t-1 == SLns_add_intrin_var_table(ns,%s_IVars,NULL)  ||\n",mname);
	if (iface.intrin_arrays != NULL)
	   emit("\t-1 == make_intrinsic_arrays(ns)	||\n",mname);
   }

   emit("\t-1 == SLns_add_intrin_fun_table (ns,%s_Funcs,(char*)\"__%s__\"))\n"+
					"\t\treturn -1;\n", mname, mname);

   emit("\n   return 0;\n}\n");
   emit("END_DECLS\n");
}

define generate_opaque_code()
{
   !if (length(SC.opaque_types)) return 0;

   variable non_reserved = SC.opaque_types[ [SC.num_reserved_types:]];
   variable external = struct_map(Integer_Type, non_reserved, "external");
   variable internal = non_reserved[ where (not(external)) ];
   external = non_reserved[ where(external) ];
   variable tmap, num_new_opaques = length(internal);

   if (length(external)) {
	% These declarations permit compilation of one module when 
	% generated against the typemaps of another module.
	foreach(external) {
		tmap = ();
		emit("extern SLtype %s;\n",tmap.typeid);
	}
	emit("#define SLIRP_EXTERN	extern\n");
   }
   else
	emit("#define SLIRP_EXTERN\n");

   inject_file("opaques.c");

   if (length(accepts_null_args))
	inject_file("nullargs.c");

   variable allocations = EMPTY;
   foreach (internal) {

	variable ot = (), sltype = ot.sltype;

	% Instantiate each new/unique type that will be added by this module
	emit("\nSLtype %s_Type = 0;", sltype);

 	variable parent = ot.parent;
	if (parent != NULL)
	   parent = parent.typeid;

	allocations = strcat(allocations, sprintf(
		"\n   %s_Type = allocate_opaque((char*)\"%s\","+
		"\n\t\t(FINALIZER) %S, \n\t\t(INITIALIZER) %S,\n"+
		"\t\t(SLtype) %S);"+
		"\n   if (%s_Type == SLANG_UNDEFINED_TYPE) return -1;\n",
		sltype, sltype, ot.finalizer, ot.initializer, parent, sltype));
   }

   if (num_new_opaques) {

	allocations = strcat(

		sprintf("   if (%s_Type) return 0;\n\n", internal[0].sltype),

		sprintf("   Slirp_Opaque_Types = (Slirp_Type**) "+
			"SLrealloc((char*)Slirp_Opaque_Types,\n\t\t(1 + "+
			"Slirp_Num_Opaque_Types + %d) * sizeof(Slirp_Type*));"+
			"\n\n", num_new_opaques),

		"   if (Slirp_Opaque_Types == NULL) return -1;\n",

		allocations);

	emit("\n\nstatic int allocate_%s_opaque_types(void)\n",SC.modname);
	emit("{\n%s\n   return 0;\n}\n",allocations);
   }

   return num_new_opaques;
}
% }}}

static define emit_assoc_array(array, format) % {{{
{
   foreach(array) using ("keys") {
	variable elem = ();
	emit(format, elem);
   }
} % }}}

define generate_makefile() % {{{
{
   SC.gluefp = create_output_file("", ".mf");
   emit("MODNAME = %s\n", SC.modname);

   emit("WRAPPED_CODE = ");
   emit_assoc_array(SC.lib_paths, " -L%s");
   emit_assoc_array(SC.lib_names, " -l%s");
   emit_assoc_array(SC.ldflags, " %s");

   emit("\nWRAPPED_INCS = ");
   emit_assoc_array(SC.header_paths, " -I%s");
   if (SC.fortran)
	emit(" $(F2C_INCLUDE)");

   % Use original FORTRAN source as the input, not f2c-generated .h
   emit("\nMISC_LIBS = ");
   variable input = @SC.header_names;
   if (SC.fortran) {
	emit("$(F2C_LIB)");
	variable i = length(input);
	while(i,i--) { (input[i], ) = strreplace(input[i], ".f.h", ".f", -1); }
   }

   emit("\nWRAPPED_HDRS = ");
   emit( strjoin(input, " "));
   emit("\n\nLOCAL_CODE = \n");
   emit("SLIRP = %s %s -m %s", __process_name, __argv[0], SC.modname);
   emit("\n\n");

   inject_file("unix.mf");
   close_file(SC.gluefp);

   dprintf("Starter make file generated to '%s'", SC.modname + ".mf");
   dprintf("Type 'make -f %s' to build module", SC.modname + ".mf");
   dprintf("You may need to set WRAPPED_CODE or other macros before linking.");
}
% }}}

% Language handlers {{{

static define c_and_cpp(file) { SC.fortran = 0; return file; }
_handlers[".h"]		= &c_and_cpp;		% default handler
_handlers[".hh"]	= &c_and_cpp;

#ifeval SC.have_f2c
static define f2c(in)
{
   variable out = sprintf("%s.h",path_concat(SC.gendir,path_basename(in)));
   () = system(sprintf("%s %s %s",path_concat(SC.root,"f2cprotos"),out,in));

   if (stat_file(out) != NULL) {
	SC.have_fortran = 1;
	set_wrapper_semantics(1, 1);

	% Convenience:  if -make specified but no library args are given,
	% 		automatically add .o file to link line
	if (andelse {SC.genmakef} {length(SC.ldflags) == 0}
				  {length(SC.lib_names) == 0})
	   SC.ldflags[ path_sans_extname(in) + ".o" ] = out;

	return out;
   }

   if (stat_file(out + ".err") != NULL)
	() = system("cat " + out+ ".err");

   return NULL;
}
_handlers[".f"]		= &f2c;
#endif
% }}}

provide("slirputils");
